@@ -1,7 +1,7 @@
 package genswagger
 
 import (
-	"encoding/json"
+	"fmt"
 	"reflect"
 	"testing"
 
@@ -253,32 +253,26 @@ func TestApplyTemplateSimple(t *testing.T) {
 		t.Errorf("applyTemplate(%#v) failed with %v; want success", file, err)
 		return
 	}
-	got := new(swaggerObject)
-	err = json.Unmarshal([]byte(result), got)
-	if err != nil {
-		t.Errorf("json.Unmarshal(%s) failed with %v; want success", result, err)
-		return
-	}
-	if want, is, name := "2.0", got.Swagger, "Swagger"; !reflect.DeepEqual(is, want) {
+	if want, is, name := "2.0", result.Swagger, "Swagger"; !reflect.DeepEqual(is, want) {
 		t.Errorf("applyTemplate(%#v).%s = %s want to be %s", file, name, is, want)
 	}
-	if want, is, name := "", got.BasePath, "BasePath"; !reflect.DeepEqual(is, want) {
+	if want, is, name := "", result.BasePath, "BasePath"; !reflect.DeepEqual(is, want) {
 		t.Errorf("applyTemplate(%#v).%s = %s want to be %s", file, name, is, want)
 	}
-	if want, is, name := []string{"http", "https"}, got.Schemes, "Schemes"; !reflect.DeepEqual(is, want) {
+	if want, is, name := []string{"http", "https"}, result.Schemes, "Schemes"; !reflect.DeepEqual(is, want) {
 		t.Errorf("applyTemplate(%#v).%s = %s want to be %s", file, name, is, want)
 	}
-	if want, is, name := []string{"application/json"}, got.Consumes, "Consumes"; !reflect.DeepEqual(is, want) {
+	if want, is, name := []string{"application/json"}, result.Consumes, "Consumes"; !reflect.DeepEqual(is, want) {
 		t.Errorf("applyTemplate(%#v).%s = %s want to be %s", file, name, is, want)
 	}
-	if want, is, name := []string{"application/json"}, got.Produces, "Produces"; !reflect.DeepEqual(is, want) {
+	if want, is, name := []string{"application/json"}, result.Produces, "Produces"; !reflect.DeepEqual(is, want) {
 		t.Errorf("applyTemplate(%#v).%s = %s want to be %s", file, name, is, want)
 	}
 
 	// If there was a failure, print out the input and the json result for debugging.
 	if t.Failed() {
 		t.Errorf("had: %s", file)
-		t.Errorf("got: %s", result)
+		t.Errorf("got: %s", fmt.Sprint(result))
 	}
 }
 
@@ -413,35 +407,29 @@ func TestApplyTemplateRequestWithoutClientStreaming(t *testing.T) {
 		t.Errorf("applyTemplate(%#v) failed with %v; want success", file, err)
 		return
 	}
-	var obj swaggerObject
-	err = json.Unmarshal([]byte(result), &obj)
-	if err != nil {
-		t.Errorf("applyTemplate(%#v) failed with %v; want success", file, err)
-		return
-	}
-	if want, got := "2.0", obj.Swagger; !reflect.DeepEqual(got, want) {
+	if want, got := "2.0", result.Swagger; !reflect.DeepEqual(got, want) {
 		t.Errorf("applyTemplate(%#v).Swagger = %s want to be %s", file, got, want)
 	}
-	if want, got := "", obj.BasePath; !reflect.DeepEqual(got, want) {
+	if want, got := "", result.BasePath; !reflect.DeepEqual(got, want) {
 		t.Errorf("applyTemplate(%#v).BasePath = %s want to be %s", file, got, want)
 	}
-	if want, got := []string{"http", "https"}, obj.Schemes; !reflect.DeepEqual(got, want) {
+	if want, got := []string{"http", "https"}, result.Schemes; !reflect.DeepEqual(got, want) {
 		t.Errorf("applyTemplate(%#v).Schemes = %s want to be %s", file, got, want)
 	}
-	if want, got := []string{"application/json"}, obj.Consumes; !reflect.DeepEqual(got, want) {
+	if want, got := []string{"application/json"}, result.Consumes; !reflect.DeepEqual(got, want) {
 		t.Errorf("applyTemplate(%#v).Consumes = %s want to be %s", file, got, want)
 	}
-	if want, got := []string{"application/json"}, obj.Produces; !reflect.DeepEqual(got, want) {
+	if want, got := []string{"application/json"}, result.Produces; !reflect.DeepEqual(got, want) {
 		t.Errorf("applyTemplate(%#v).Produces = %s want to be %s", file, got, want)
 	}
-	if want, got, name := "Generated for ExampleService.Echo - ", obj.Paths["/v1/echo"].Post.Summary, "Paths[/v1/echo].Post.Summary"; !reflect.DeepEqual(got, want) {
+	if want, got, name := "Generated for ExampleService.Echo - ", result.Paths["/v1/echo"].Post.Summary, "Paths[/v1/echo].Post.Summary"; !reflect.DeepEqual(got, want) {
 		t.Errorf("applyTemplate(%#v).%s = %s want to be %s", file, name, got, want)
 	}
 
 	// If there was a failure, print out the input and the json result for debugging.
 	if t.Failed() {
 		t.Errorf("had: %s", file)
-		t.Errorf("got: %s", result)
+		t.Errorf("got: %s", fmt.Sprint(result))
 	}
 }
 
@@ -685,22 +673,16 @@ func TestApplyTemplateRequestWithUnusedReferences(t *testing.T) {
 		t.Errorf("applyTemplate(%#v) failed with %v; want success", file, err)
 		return
 	}
-	var obj swaggerObject
-	err = json.Unmarshal([]byte(result), &obj)
-	if err != nil {
-		t.Errorf("applyTemplate(%#v) failed with %v; want success", file, err)
-		return
-	}
 
 	// Only EmptyMessage must be present, not ExampleMessage
-	if want, got, name := 1, len(obj.Definitions), "len(Definitions)"; !reflect.DeepEqual(got, want) {
+	if want, got, name := 1, len(result.Definitions), "len(Definitions)"; !reflect.DeepEqual(got, want) {
 		t.Errorf("applyTemplate(%#v).%s = %d want to be %d", file, name, got, want)
 	}
 
 	// If there was a failure, print out the input and the json result for debugging.
 	if t.Failed() {
 		t.Errorf("had: %s", file)
-		t.Errorf("got: %s", result)
+		t.Errorf("got: %s", fmt.Sprint(result))
 	}
 }
 
@@ -715,6 +697,16 @@ func TestTemplateToSwaggerPath(t *testing.T) {
 		{"/{test=prefix/that/has/multiple/parts/to/it/*}", "/{test}"},
 		{"/{test1}/{test2}", "/{test1}/{test2}"},
 		{"/{test1}/{test2}/", "/{test1}/{test2}/"},
+		{"/{name=prefix/*}", "/{name=prefix/*}"},
+		{"/{name=prefix1/*/prefix2/*}", "/{name=prefix1/*/prefix2/*}"},
+		{"/{user.name=prefix/*}", "/{user.name=prefix/*}"},
+		{"/{user.name=prefix1/*/prefix2/*}", "/{user.name=prefix1/*/prefix2/*}"},
+		{"/{parent=prefix/*}/children", "/{parent=prefix/*}/children"},
+		{"/{name=prefix/*}:customMethod", "/{name=prefix/*}:customMethod"},
+		{"/{name=prefix1/*/prefix2/*}:customMethod", "/{name=prefix1/*/prefix2/*}:customMethod"},
+		{"/{user.name=prefix/*}:customMethod", "/{user.name=prefix/*}:customMethod"},
+		{"/{user.name=prefix1/*/prefix2/*}:customMethod", "/{user.name=prefix1/*/prefix2/*}:customMethod"},
+		{"/{parent=prefix/*}/children:customMethod", "/{parent=prefix/*}/children:customMethod"},
 	}
 
 	for _, data := range tests {
@@ -787,3 +779,385 @@ func TestFQMNtoSwaggerName(t *testing.T) {
 		}
 	}
 }
+
+func TestSchemaOfField(t *testing.T) {
+	type test struct {
+		field    *descriptor.Field
+		refs     refMap
+		expected schemaCore
+	}
+
+	tests := []test{
+		{
+			field: &descriptor.Field{
+				FieldDescriptorProto: &protodescriptor.FieldDescriptorProto{
+					Name: proto.String("primitive_field"),
+					Type: protodescriptor.FieldDescriptorProto_TYPE_STRING.Enum(),
+				},
+			},
+			refs: make(refMap),
+			expected: schemaCore{
+				Type: "string",
+			},
+		},
+		{
+			field: &descriptor.Field{
+				FieldDescriptorProto: &protodescriptor.FieldDescriptorProto{
+					Name:  proto.String("repeated_primitive_field"),
+					Type:  protodescriptor.FieldDescriptorProto_TYPE_STRING.Enum(),
+					Label: protodescriptor.FieldDescriptorProto_LABEL_REPEATED.Enum(),
+				},
+			},
+			refs: make(refMap),
+			expected: schemaCore{
+				Type: "array",
+				Items: &swaggerItemsObject{
+					Type: "string",
+				},
+			},
+		},
+		{
+			field: &descriptor.Field{
+				FieldDescriptorProto: &protodescriptor.FieldDescriptorProto{
+					Name:     proto.String("wrapped_field"),
+					TypeName: proto.String(".google.protobuf.StringValue"),
+					Type:     protodescriptor.FieldDescriptorProto_TYPE_MESSAGE.Enum(),
+				},
+			},
+			refs: make(refMap),
+			expected: schemaCore{
+				Type: "string",
+			},
+		},
+		{
+			field: &descriptor.Field{
+				FieldDescriptorProto: &protodescriptor.FieldDescriptorProto{
+					Name:     proto.String("repeated_wrapped_field"),
+					TypeName: proto.String(".google.protobuf.StringValue"),
+					Type:     protodescriptor.FieldDescriptorProto_TYPE_MESSAGE.Enum(),
+					Label:    protodescriptor.FieldDescriptorProto_LABEL_REPEATED.Enum(),
+				},
+			},
+			refs: make(refMap),
+			expected: schemaCore{
+				Type: "array",
+				Items: &swaggerItemsObject{
+					Type: "string",
+				},
+			},
+		},
+		{
+			field: &descriptor.Field{
+				FieldDescriptorProto: &protodescriptor.FieldDescriptorProto{
+					Name:     proto.String("wrapped_field"),
+					TypeName: proto.String(".google.protobuf.BytesValue"),
+					Type:     protodescriptor.FieldDescriptorProto_TYPE_MESSAGE.Enum(),
+				},
+			},
+			refs: make(refMap),
+			expected: schemaCore{
+				Type:   "string",
+				Format: "bytes",
+			},
+		},
+		{
+			field: &descriptor.Field{
+				FieldDescriptorProto: &protodescriptor.FieldDescriptorProto{
+					Name:     proto.String("wrapped_field"),
+					TypeName: proto.String(".google.protobuf.Int32Value"),
+					Type:     protodescriptor.FieldDescriptorProto_TYPE_MESSAGE.Enum(),
+				},
+			},
+			refs: make(refMap),
+			expected: schemaCore{
+				Type:   "integer",
+				Format: "int32",
+			},
+		},
+		{
+			field: &descriptor.Field{
+				FieldDescriptorProto: &protodescriptor.FieldDescriptorProto{
+					Name:     proto.String("wrapped_field"),
+					TypeName: proto.String(".google.protobuf.UInt32Value"),
+					Type:     protodescriptor.FieldDescriptorProto_TYPE_MESSAGE.Enum(),
+				},
+			},
+			refs: make(refMap),
+			expected: schemaCore{
+				Type:   "integer",
+				Format: "int64",
+			},
+		},
+		{
+			field: &descriptor.Field{
+				FieldDescriptorProto: &protodescriptor.FieldDescriptorProto{
+					Name:     proto.String("wrapped_field"),
+					TypeName: proto.String(".google.protobuf.Int64Value"),
+					Type:     protodescriptor.FieldDescriptorProto_TYPE_MESSAGE.Enum(),
+				},
+			},
+			refs: make(refMap),
+			expected: schemaCore{
+				Type:   "string",
+				Format: "int64",
+			},
+		},
+		{
+			field: &descriptor.Field{
+				FieldDescriptorProto: &protodescriptor.FieldDescriptorProto{
+					Name:     proto.String("wrapped_field"),
+					TypeName: proto.String(".google.protobuf.UInt64Value"),
+					Type:     protodescriptor.FieldDescriptorProto_TYPE_MESSAGE.Enum(),
+				},
+			},
+			refs: make(refMap),
+			expected: schemaCore{
+				Type:   "string",
+				Format: "uint64",
+			},
+		},
+		{
+			field: &descriptor.Field{
+				FieldDescriptorProto: &protodescriptor.FieldDescriptorProto{
+					Name:     proto.String("wrapped_field"),
+					TypeName: proto.String(".google.protobuf.FloatValue"),
+					Type:     protodescriptor.FieldDescriptorProto_TYPE_MESSAGE.Enum(),
+				},
+			},
+			refs: make(refMap),
+			expected: schemaCore{
+				Type:   "number",
+				Format: "float",
+			},
+		},
+		{
+			field: &descriptor.Field{
+				FieldDescriptorProto: &protodescriptor.FieldDescriptorProto{
+					Name:     proto.String("wrapped_field"),
+					TypeName: proto.String(".google.protobuf.DoubleValue"),
+					Type:     protodescriptor.FieldDescriptorProto_TYPE_MESSAGE.Enum(),
+				},
+			},
+			refs: make(refMap),
+			expected: schemaCore{
+				Type:   "number",
+				Format: "double",
+			},
+		},
+		{
+			field: &descriptor.Field{
+				FieldDescriptorProto: &protodescriptor.FieldDescriptorProto{
+					Name:     proto.String("wrapped_field"),
+					TypeName: proto.String(".google.protobuf.BoolValue"),
+					Type:     protodescriptor.FieldDescriptorProto_TYPE_MESSAGE.Enum(),
+				},
+			},
+			refs: make(refMap),
+			expected: schemaCore{
+				Type:   "boolean",
+				Format: "boolean",
+			},
+		},
+		{
+			field: &descriptor.Field{
+				FieldDescriptorProto: &protodescriptor.FieldDescriptorProto{
+					Name:     proto.String("message_field"),
+					TypeName: proto.String(".example.Message"),
+					Type:     protodescriptor.FieldDescriptorProto_TYPE_MESSAGE.Enum(),
+				},
+			},
+			refs: refMap{".example.Message": struct{}{}},
+			expected: schemaCore{
+				Ref: "#/definitions/exampleMessage",
+			},
+		},
+	}
+
+	reg := descriptor.NewRegistry()
+	reg.Load(&plugin.CodeGeneratorRequest{
+		ProtoFile: []*protodescriptor.FileDescriptorProto{
+			{
+				SourceCodeInfo: &protodescriptor.SourceCodeInfo{},
+				Name:           proto.String("example.proto"),
+				Package:        proto.String("example"),
+				Dependency:     []string{},
+				MessageType: []*protodescriptor.DescriptorProto{
+					{
+						Name: proto.String("Message"),
+						Field: []*protodescriptor.FieldDescriptorProto{
+							{
+								Name: proto.String("value"),
+								Type: protodescriptor.FieldDescriptorProto_TYPE_STRING.Enum(),
+							},
+						},
+					},
+				},
+				EnumType: []*protodescriptor.EnumDescriptorProto{
+					{
+						Name: proto.String("Message"),
+					},
+				},
+				Service: []*protodescriptor.ServiceDescriptorProto{},
+			},
+		},
+	})
+
+	for _, test := range tests {
+		refs := make(refMap)
+		actual := schemaOfField(test.field, reg, refs)
+		if e, a := test.expected.Type, actual.Type; e != a {
+			t.Errorf("Expected schemaOfField(%v).Type = %s, actual: %s", test.field, e, a)
+		}
+		if e, a := test.expected.Ref, actual.Ref; e != a {
+			t.Errorf("Expected schemaOfField(%v).Ref = %s, actual: %s", test.field, e, a)
+		}
+		if e, a := test.expected.Items.getType(), actual.Items.getType(); e != a {
+			t.Errorf("Expected schemaOfField(%v).Items.Type = %v, actual.Type: %v", test.field, e, a)
+		}
+		if !reflect.DeepEqual(refs, test.refs) {
+			t.Errorf("Expected schemaOfField(%v) to add refs %v, not %v", test.field, test.refs, refs)
+		}
+	}
+}
+
+func fileFixtureServices(services []string) *descriptor.File {
+	var (
+		svcdesc  = []*protodescriptor.ServiceDescriptorProto{}
+		loc      = []*protodescriptor.SourceCodeInfo_Location{}
+		msgdesc  = []*protodescriptor.DescriptorProto{}
+		methdesc = []*protodescriptor.MethodDescriptorProto{}
+		msg      = []*descriptor.Message{}
+		svc      = []*descriptor.Service{}
+	)
+
+	for _, service := range services {
+		// loc
+		sc := &protodescriptor.SourceCodeInfo_Location{
+			LeadingComments: proto.String(fmt.Sprintf("%s message", service)),
+		}
+		loc = append(loc, sc)
+		// methdesc
+		md := &protodescriptor.MethodDescriptorProto{
+			Name:       proto.String(service),
+			InputType:  proto.String(fmt.Sprintf("%sMessage", service)),
+			OutputType: proto.String(fmt.Sprintf("%sMessage", service)),
+		}
+		methdesc = append(methdesc, md)
+		// svcdesc
+		sd := &protodescriptor.ServiceDescriptorProto{
+			Name:   proto.String(fmt.Sprintf("%sService", service)),
+			Method: []*protodescriptor.MethodDescriptorProto{md},
+		}
+		svcdesc = append(svcdesc, sd)
+		// msgdesc
+		d := &protodescriptor.DescriptorProto{
+			Name: proto.String(fmt.Sprintf("%sRequest", service)),
+		}
+		msgdesc = append(msgdesc, d)
+		// msg
+		m := &descriptor.Message{
+			DescriptorProto: d,
+		}
+		msg = append(msg, m)
+		// methods
+		meth := &descriptor.Method{
+			MethodDescriptorProto: md,
+			RequestType:           m,
+			ResponseType:          m,
+			Bindings: []*descriptor.Binding{
+				{
+					HTTPMethod: "POST",
+					Body:       &descriptor.Body{FieldPath: nil},
+					PathTmpl: httprule.Template{
+						Version:  1,
+						OpCodes:  []int{0, 0},
+						Template: fmt.Sprintf("/v1/%s", service),
+					},
+				},
+			},
+		}
+		s := &descriptor.Service{
+			ServiceDescriptorProto: sd,
+			Methods:                []*descriptor.Method{meth},
+		}
+		svc = append(svc, s)
+	}
+
+	file := descriptor.File{
+		FileDescriptorProto: &protodescriptor.FileDescriptorProto{
+			SourceCodeInfo: &protodescriptor.SourceCodeInfo{
+				Location: loc,
+			},
+			Name:        proto.String("example.proto"),
+			Package:     proto.String("example"),
+			Dependency:  []string{},
+			MessageType: msgdesc,
+			Service:     svcdesc,
+		},
+		GoPkg: descriptor.GoPackage{
+			Path: "example.com/path/to/example/example1.pb",
+			Name: "example_pb",
+		},
+		Messages: msg,
+		Services: svc,
+	}
+	return &file
+}
+
+func TestIsProtoPathMatches(t *testing.T) {
+	for _, test := range []struct {
+		path       []int32
+		outerPaths []int32
+		typeName   string
+		typeIndex  int32
+		fieldPaths []int32
+		want       bool
+	}{
+		{[]int32{6, 0, 2, 0}, []int32{}, "Method", 1, []int32{2, 0}, false},
+		{[]int32{6, 0, 2, 0, 1}, []int32{}, "Method", 1, []int32{2, 0}, false},
+		{[]int32{6, 0, 2, 0}, []int32{}, "Method", 0, []int32{2, 0}, true},
+		{[]int32{}, []int32{}, "Package", 2, []int32{}, false},
+		{[]int32{2}, []int32{}, "Package", 3, []int32{}, false},
+		{[]int32{2}, []int32{}, "Package", 2, []int32{}, true},
+		{[]int32{4, 0}, []int32{}, "MessageType", 1, []int32{}, false},
+		{[]int32{4, 0, 2, 0}, []int32{}, "MessageType", 0, []int32{4, 0}, false},
+	} {
+		got := isProtoPathMatches(test.path, test.outerPaths, test.typeName, test.typeIndex, test.fieldPaths)
+		if got != test.want {
+			t.Errorf("isProtoPathMatches(%v) got (%t) want %t", test, got, test.want)
+		}
+	}
+}
+
+func TestProtoComments(t *testing.T) {
+	for _, test := range []struct {
+		services   []string
+		startPath  []int32
+		outers     []string
+		typeName   string
+		typeIndex  int32
+		fieldPaths []int32
+		want       string
+	}{
+		{[]string{"Foo", "Bar"}, []int32{6, 0, 2, 0}, []string{}, "Method", 0, []int32{2, 0}, "Foo message"},
+		{[]string{"Foo", "Bar"}, []int32{6, 0, 2, 0}, []string{}, "Method", 1, []int32{2, 0}, "Bar message"},
+		{[]string{"Foo", "Bar"}, []int32{6, 0, 2, 0}, []string{}, "Method", 2, []int32{2, 0}, ""},
+		{[]string{"Foo", "Bar"}, []int32{4, 0, 1}, []string{}, "Method", 0, []int32{2, 0}, ""},
+		{[]string{"Foo", "Bar"}, []int32{6, 0, 1, 1}, []string{}, "Method", 0, []int32{2, 0}, ""},
+		{[]string{"Foo", "Bar", "Baz"}, []int32{6, 0, 2, 0}, []string{}, "Method", 2, []int32{2, 0}, "Baz message"},
+	} {
+		file := fileFixtureServices(test.services)
+		for i, loc := range file.SourceCodeInfo.Location {
+			test.startPath[1] = int32(i)
+			loc.Path = append(loc.Path, test.startPath...)
+		}
+		reg := descriptor.NewRegistry()
+		reg.Load(&plugin.CodeGeneratorRequest{
+			ProtoFile: []*protodescriptor.FileDescriptorProto{file.FileDescriptorProto},
+		})
+		got := protoComments(reg, file, test.outers, test.typeName, test.typeIndex, test.fieldPaths...)
+		if got != test.want {
+			t.Errorf("protoComments(%v) got (%s) want %s", test, got, test.want)
+		}
+	}
+}
